
// Array remote access and transformational functions.


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// `synch'                                                             //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

inline void synch() {
  TreeSchedule spanTree ;
  :: spanTree(&spanTree, apg) ;

  spanTree.synch() ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// `get'                                                               //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
inline void get(T* base, const SectionRM0<T>& a) {
  Get schedule(Seq0<int>(), base, &a, a.offset, a.ptrs, sizeof(T)) ;

  schedule.gather() ;
}

template<class T>
inline void get(const int str_0, T* base, const SectionRM1<T>& a) {
  Get schedule(Seq1<int>(str_0), base, &a, a.offset, a.ptrs, sizeof(T)) ;

  schedule.gather() ;
}

template<class T>
inline void get(const int str_0, 
                const int str_1, T* base, const SectionRM2<T>& a) {
  Get schedule(Seq2<int>(str_0, str_1), base, &a, a.offset, a.ptrs, sizeof(T)) ;

  schedule.gather() ;
}

template<class T>
inline void get(const int str_0, 
                const int str_1, 
                const int str_2, T* base, const SectionRM3<T>& a) {
  Get schedule(Seq3<int>(str_0, str_1, str_2), base, &a, a.offset, a.ptrs, sizeof(T)) ;

  schedule.gather() ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// `remap' and `writeHalo'                                             //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
inline void remapRM(const Section0<T>& b, const SectionRM0<T>& a) {
  RemapRM schedule(&b, b.dat(),
                   &a, a.offset, a.ptrs,
                   sizeof(T)) ;

  schedule.execute() ;
}

template<class T>
inline void remapRM(const Section1<T>& b, const SectionRM1<T>& a) {
  RemapRM schedule(&b, b.dat(),
                   &a, a.offset, a.ptrs,
                   sizeof(T)) ;

  schedule.execute() ;
}

template<class T>
inline void remapRM(const Section2<T>& b, const SectionRM2<T>& a) {
  RemapRM schedule(&b, b.dat(),
                   &a, a.offset, a.ptrs,
                   sizeof(T)) ;

  schedule.execute() ;
}

template<class T>
inline void remapRM(const Section3<T>& b, const SectionRM3<T>& a) {
  RemapRM schedule(&b, b.dat(),
                   &a, a.offset, a.ptrs,
                   sizeof(T)) ;

  schedule.execute() ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// `gather', `scatter', `vecGather' and `vecScatter'.                  //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

// Simple gather

template<class T>
inline void gatherRM(const Section1<T>& b, const SectionRM1<T>& a,
                     const Section1<int>& s_0) {
  GatherRM schedule(&b, b.dat(),
                    &a, a.offset, a.ptrs,
                    sizeof(T),
                    Seq1<const DAD*>(&s_0),
                    Seq1<const int*>(s_0.dat()),
                    0, 0) ;

  schedule.execute() ;
}

template<class T>
inline void gatherRM(const Section2<T>& b, const SectionRM1<T>& a,
                     const Section2<int>& s_0) {
  GatherRM schedule(&b, b.dat(),
                    &a, a.offset, a.ptrs,
                    sizeof(T),
                    Seq1<const DAD*>(&s_0),
                    Seq1<const int*>(s_0.dat()),
                    0, 0) ;

  schedule.execute() ;
}

template<class T>
inline void gatherRM(const Section3<T>& b, const SectionRM1<T>& a,
                     const Section3<int>& s_0) {
  GatherRM schedule(&b, b.dat(),
                    &a, a.offset, a.ptrs,
                    sizeof(T),
                    Seq1<const DAD*>(&s_0),
                    Seq1<const int*>(s_0.dat()),
                    0, 0) ;

  schedule.execute() ;
}

template<class T>
inline void gatherRM(const Section1<T>& b, const SectionRM2<T>& a,
                     const Section1<int>& s_0,
                     const Section1<int>& s_1) {
  GatherRM schedule(&b, b.dat(),
                    &a, a.offset, a.ptrs,
                    sizeof(T),
                    Seq2<const DAD*>(&s_0, &s_1),
                    Seq2<const int*>(s_0.dat(), s_1.dat()),
                    0, 0) ;

  schedule.execute() ;
}

template<class T>
inline void gatherRM(const Section2<T>& b, const SectionRM2<T>& a,
                     const Section2<int>& s_0,
                     const Section2<int>& s_1) {
  GatherRM schedule(&b, b.dat(),
                    &a, a.offset, a.ptrs,
                    sizeof(T),
                    Seq2<const DAD*>(&s_0, &s_1),
                    Seq2<const int*>(s_0.dat(), s_1.dat()),
                    0, 0) ;

  schedule.execute() ;
}

template<class T>
inline void gatherRM(const Section3<T>& b, const SectionRM2<T>& a,
                     const Section3<int>& s_0,
                     const Section3<int>& s_1) {
  GatherRM schedule(&b, b.dat(),
                    &a, a.offset, a.ptrs,
                    sizeof(T),
                    Seq2<const DAD*>(&s_0, &s_1),
                    Seq2<const int*>(s_0.dat(), s_1.dat()),
                    0, 0) ;

  schedule.execute() ;
}

template<class T>
inline void gatherRM(const Section1<T>& b, const SectionRM3<T>& a,
                     const Section1<int>& s_0,
                     const Section1<int>& s_1,
                     const Section1<int>& s_2) {
  GatherRM schedule(&b, b.dat(),
                    &a, a.offset, a.ptrs,
                    sizeof(T),
                    Seq3<const DAD*>(&s_0, &s_1, &s_2),
                    Seq3<const int*>(s_0.dat(), s_1.dat(), s_2.dat()),
                    0, 0) ;

  schedule.execute() ;
}

template<class T>
inline void gatherRM(const Section2<T>& b, const SectionRM3<T>& a,
                     const Section2<int>& s_0,
                     const Section2<int>& s_1,
                     const Section2<int>& s_2) {
  GatherRM schedule(&b, b.dat(),
                    &a, a.offset, a.ptrs,
                    sizeof(T),
                    Seq3<const DAD*>(&s_0, &s_1, &s_2),
                    Seq3<const int*>(s_0.dat(), s_1.dat(), s_2.dat()),
                    0, 0) ;

  schedule.execute() ;
}

template<class T>
inline void gatherRM(const Section3<T>& b, const SectionRM3<T>& a,
                     const Section3<int>& s_0,
                     const Section3<int>& s_1,
                     const Section3<int>& s_2) {
  GatherRM schedule(&b, b.dat(),
                    &a, a.offset, a.ptrs,
                    sizeof(T),
                    Seq3<const DAD*>(&s_0, &s_1, &s_2),
                    Seq3<const int*>(s_0.dat(), s_1.dat(), s_2.dat()),
                    0, 0) ;

  schedule.execute() ;
}


// Masked gather

template<class T>
inline void gatherRM(const Section1<T>& b, const SectionRM1<T>& a,
                     const Section1<int>& s_0,
                     const Section1<int>& msk) {
  GatherRM schedule(&b, b.dat(),
                    &a, a.offset, a.ptrs,
                    sizeof(T),
                    Seq1<const DAD*>(&s_0),
                    Seq1<const int*>(s_0.dat()),
                    &msk, msk.dat()) ;

  schedule.execute() ;
}

template<class T>
inline void gatherRM(const Section2<T>& b, const SectionRM1<T>& a,
                     const Section2<int>& s_0,
                     const Section2<int>& msk) {
  GatherRM schedule(&b, b.dat(),
                    &a, a.offset, a.ptrs,
                    sizeof(T),
                    Seq1<const DAD*>(&s_0),
                    Seq1<const int*>(s_0.dat()),
                    &msk, msk.dat()) ;

  schedule.execute() ;
}

template<class T>
inline void gatherRM(const Section3<T>& b, const SectionRM1<T>& a,
                     const Section3<int>& s_0,
                     const Section3<int>& msk) {
  GatherRM schedule(&b, b.dat(),
                    &a, a.offset, a.ptrs,
                    sizeof(T),
                    Seq1<const DAD*>(&s_0),
                    Seq1<const int*>(s_0.dat()),
                    &msk, msk.dat()) ;

  schedule.execute() ;
}

template<class T>
inline void gatherRM(const Section1<T>& b, const SectionRM2<T>& a,
                     const Section1<int>& s_0,
                     const Section1<int>& s_1,
                     const Section1<int>& msk) {
  GatherRM schedule(&b, b.dat(),
                    &a, a.offset, a.ptrs,
                    sizeof(T),
                    Seq2<const DAD*>(&s_0, &s_1),
                    Seq2<const int*>(s_0.dat(), s_1.dat()),
                    &msk, msk.dat()) ;

  schedule.execute() ;
}

template<class T>
inline void gatherRM(const Section2<T>& b, const SectionRM2<T>& a,
                     const Section2<int>& s_0,
                     const Section2<int>& s_1,
                     const Section2<int>& msk) {
  GatherRM schedule(&b, b.dat(),
                    &a, a.offset, a.ptrs,
                    sizeof(T),
                    Seq2<const DAD*>(&s_0, &s_1),
                    Seq2<const int*>(s_0.dat(), s_1.dat()),
                    &msk, msk.dat()) ;

  schedule.execute() ;
}

template<class T>
inline void gatherRM(const Section3<T>& b, const SectionRM2<T>& a,
                     const Section3<int>& s_0,
                     const Section3<int>& s_1,
                     const Section3<int>& msk) {
  GatherRM schedule(&b, b.dat(),
                    &a, a.offset, a.ptrs,
                    sizeof(T),
                    Seq2<const DAD*>(&s_0, &s_1),
                    Seq2<const int*>(s_0.dat(), s_1.dat()),
                    &msk, msk.dat()) ;

  schedule.execute() ;
}

template<class T>
inline void gatherRM(const Section1<T>& b, const SectionRM3<T>& a,
                     const Section1<int>& s_0,
                     const Section1<int>& s_1,
                     const Section1<int>& s_2,
                     const Section1<int>& msk) {
  GatherRM schedule(&b, b.dat(),
                    &a, a.offset, a.ptrs,
                    sizeof(T),
                    Seq3<const DAD*>(&s_0, &s_1, &s_2),
                    Seq3<const int*>(s_0.dat(), s_1.dat(), s_2.dat()),
                    &msk, msk.dat()) ;

  schedule.execute() ;
}

template<class T>
inline void gatherRM(const Section2<T>& b, const SectionRM3<T>& a,
                     const Section2<int>& s_0,
                     const Section2<int>& s_1,
                     const Section2<int>& s_2,
                     const Section2<int>& msk) {
  GatherRM schedule(&b, b.dat(),
                    &a, a.offset, a.ptrs,
                    sizeof(T),
                    Seq3<const DAD*>(&s_0, &s_1, &s_2),
                    Seq3<const int*>(s_0.dat(), s_1.dat(), s_2.dat()),
                    &msk, msk.dat()) ;

  schedule.execute() ;
}

template<class T>
inline void gatherRM(const Section3<T>& b, const SectionRM3<T>& a,
                     const Section3<int>& s_0,
                     const Section3<int>& s_1,
                     const Section3<int>& s_2,
                     const Section3<int>& msk) {
  GatherRM schedule(&b, b.dat(),
                    &a, a.offset, a.ptrs,
                    sizeof(T),
                    Seq3<const DAD*>(&s_0, &s_1, &s_2),
                    Seq3<const int*>(s_0.dat(), s_1.dat(), s_2.dat()),
                    &msk, msk.dat()) ;

  schedule.execute() ;
}


// Vector subscript gather.

template<class T>
inline void vecGatherRM(const Section1<T>& b, const SectionRM1<T>& a,
                        const Section1<int>* s_0) {
  VecGatherRM schedule(&b, b.dat(),
                       &a, a.offset, a.ptrs,
                       sizeof(T),
                       Seq1<const DAD*>(s_0),
                       Seq1<const int*>(s_0 ? s_0->dat() : 0)) ;

  schedule.execute() ;
}

template<class T>
inline void vecGatherRM(const Section2<T>& b, const SectionRM2<T>& a,
                        const Section1<int>* s_0,
                        const Section1<int>* s_1) {
  VecGatherRM schedule(&b, b.dat(),
                       &a, a.offset, a.ptrs,
                       sizeof(T),
                       Seq2<const DAD*>(s_0, s_1),
                       Seq2<const int*>(s_0 ? s_0->dat() : 0,
                                        s_1 ? s_1->dat() : 0)) ;

  schedule.execute() ;
}

template<class T>
inline void vecGatherRM(const Section3<T>& b, const SectionRM3<T>& a,
                        const Section1<int>* s_0,
                        const Section1<int>* s_1,
                        const Section1<int>* s_2) {
  VecGatherRM schedule(&b, b.dat(),
                       &a, a.offset, a.ptrs,
                       sizeof(T),
                       Seq3<const DAD*>(s_0, s_1, s_2),
                       Seq3<const int*>(s_0 ? s_0->dat() : 0,
                                        s_1 ? s_1->dat() : 0,
                                        s_2 ? s_2->dat() : 0)) ;

  schedule.execute() ;
}


// Vector subscript scatter.

template<class T>
inline void vecScatterRM(const Section1<T>& b, const SectionRM1<T>& a,
                         const Section1<int>* s_0) {
  VecScatterRM schedule(&b, b.dat(),
                        &a, a.offset, a.ptrs,
                        sizeof(T),
                        Seq1<const DAD*>(s_0),
                        Seq1<const int*>(s_0 ? s_0->dat() : 0)) ;

  schedule.execute() ;
}

template<class T>
inline void vecScatterRM(const Section2<T>& b, const SectionRM2<T>& a,
                         const Section1<int>* s_0,
                         const Section1<int>* s_1) {
  VecScatterRM schedule(&b, b.dat(),
                        &a, a.offset, a.ptrs,
                        sizeof(T),
                        Seq2<const DAD*>(s_0, s_1),
                        Seq2<const int*>(s_0 ? s_0->dat() : 0,
                                         s_1 ? s_1->dat() : 0)) ;

  schedule.execute() ;
}

template<class T>
inline void vecScatterRM(const Section3<T>& b, const SectionRM3<T>& a,
                         const Section1<int>* s_0,
                         const Section1<int>* s_1,
                         const Section1<int>* s_2) {
  VecScatterRM schedule(&b, b.dat(),
                        &a, a.offset, a.ptrs,
                        sizeof(T),
                        Seq3<const DAD*>(s_0, s_1, s_2),
                        Seq3<const int*>(s_0 ? s_0->dat() : 0,
                                         s_1 ? s_1->dat() : 0,
                                         s_2 ? s_2->dat() : 0)) ;

  schedule.execute() ;
}

